"""
Copyright (C) Cross The Road Electronics.  All rights reserved.
License information can be found in CTRE_LICENSE.txt
For support and suggestions contact support@ctr-electronics.com or file
an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
"""

try:
    import math
    from wpimath.kinematics import (
        SwerveDrive2Kinematics,
        SwerveDrive3Kinematics,
        SwerveDrive4Kinematics,
        SwerveDrive6Kinematics,
    )
    from wpilib.simulation import DCMotorSim
    from wpimath.system.plant import DCMotor, LinearSystemId

    from phoenix6.sim.chassis_reference import ChassisReference
    from phoenix6.sim.pigeon2_sim_state import Pigeon2SimState
    from phoenix6.swerve.swerve_module_constants import SwerveModuleConstants
    from phoenix6.swerve.swerve_module import SwerveModule
    from phoenix6.swerve.utility.geometry import Rotation2d, Translation2d
    from phoenix6.swerve.utility.kinematics import SwerveModuleState
    from phoenix6.units import kilogram_square_meter, radians_per_second, second, volt

    class SimSwerveDrivetrain:
        """
        Simplified swerve drive simulation class.

        This class assumes that the swerve drive is perfect, meaning
        that there is no scrub and the wheels do not slip.

        In addition, it assumes the inertia of the robot is governed only
        by the inertia of the steer module and the individual drive wheels.
        Robot-wide inertia is not accounted for, and neither is translational
        vs rotational inertia of the robot.

        These assumptions provide a simplified example that can demonstrate the
        behavior of a swerve drive in simulation. Users are encouraged to
        expand this model for their own use.
        """

        class SimSwerveModule:
            def __init__(
                self,
                drive_gearing: float, drive_inertia: kilogram_square_meter,
                drive_friction_voltage: volt, drive_motor_inverted: bool,
                steer_gearing: float, steer_inertia: kilogram_square_meter,
                steer_friction_voltage: volt, steer_motor_inverted: bool,
                cancoder_inverted: bool,
            ):
                gearbox = DCMotor.krakenX60FOC(1)

                self.drive_motor = DCMotorSim(LinearSystemId.DCMotorSystem(gearbox, drive_inertia, drive_gearing), gearbox)
                """Reference to motor simulation for drive motor"""
                self.steer_motor = DCMotorSim(LinearSystemId.DCMotorSystem(gearbox, steer_inertia, steer_gearing), gearbox)
                """Reference to motor simulation for the steer motor"""
                self.drive_gearing = drive_gearing
                """Reference to steer gearing for updating CANcoder"""
                self.steer_gearing = steer_gearing
                """Reference to steer gearing for updating CANcoder"""
                self.drive_friction_voltage = drive_friction_voltage
                """Voltage necessary for the drive motor to overcome friction"""
                self.steer_friction_voltage = steer_friction_voltage
                """Voltage necessary for the steer motor to overcome friction"""
                self.drive_motor_inverted = drive_motor_inverted
                """Whether the drive motor is inverted"""
                self.steer_motor_inverted = steer_motor_inverted
                """Whether the steer motor is inverted"""
                self.cancoder_inverted = cancoder_inverted
                """Whether the CANcoder is inverted"""

        def __init__(
            self,
            wheel_locations: list[Translation2d],
            pigeon_sim: Pigeon2SimState,
            module_constants: list[SwerveModuleConstants],
        ):
            self._pigeon_sim = pigeon_sim
            self._modules: list[SimSwerveDrivetrain.SimSwerveModule] = []
            for module in module_constants:
                self._modules.append(
                    self.SimSwerveModule(
                        module.drive_motor_gear_ratio, module.drive_inertia,
                        module.drive_friction_voltage, module.drive_motor_inverted,
                        module.steer_motor_gear_ratio, module.steer_inertia,
                        module.steer_friction_voltage, module.steer_motor_inverted,
                        module.cancoder_inverted,
                    )
                )

            if len(wheel_locations) == 2:
                self._kinem = SwerveDrive2Kinematics(wheel_locations[0], wheel_locations[1])
            elif len(wheel_locations) == 3:
                self._kinem = SwerveDrive3Kinematics(wheel_locations[0], wheel_locations[1], wheel_locations[2])
            elif len(wheel_locations) == 4:
                self._kinem = SwerveDrive4Kinematics(wheel_locations[0], wheel_locations[1], wheel_locations[2], wheel_locations[3])
            elif len(wheel_locations) == 6:
                self._kinem = SwerveDrive6Kinematics(wheel_locations[0], wheel_locations[1], wheel_locations[2], wheel_locations[3], wheel_locations[4], wheel_locations[5])
            else:
                self._kinem = None

            self._last_angle = Rotation2d()

        def update(self, dt: second, supply_voltage: volt, modules_to_apply: list[SwerveModule]):
            """
            Update this simulation for the time duration.

            This performs a simulation update on all the simulated devices

            :param dt: The time delta between this update and the previous update
            :type dt: second
            :param supply_voltage: The voltage as seen at the motor controllers
            :type supply_voltage: volt
            :param modules_to_apply: What modules to apply the update to
            :type modules_to_apply: list[SwerveModule]
            """
            if self._kinem is None:
                return
            elif len(modules_to_apply) != len(self._modules):
                return

            for sim_module, module_to_apply in zip(self._modules, modules_to_apply):
                drive_motor = module_to_apply.drive_motor.sim_state
                steer_motor = module_to_apply.steer_motor.sim_state
                cancoder = module_to_apply.cancoder.sim_state

                drive_motor.orientation = (
                    ChassisReference.Clockwise_Positive
                    if sim_module.drive_motor_inverted
                    else ChassisReference.CounterClockwise_Positive
                )
                steer_motor.orientation = (
                    ChassisReference.Clockwise_Positive
                    if sim_module.steer_motor_inverted
                    else ChassisReference.CounterClockwise_Positive
                )
                cancoder.orientation = (
                    ChassisReference.Clockwise_Positive
                    if sim_module.cancoder_inverted
                    else ChassisReference.CounterClockwise_Positive
                )

                drive_motor.set_supply_voltage(supply_voltage)
                steer_motor.set_supply_voltage(supply_voltage)
                cancoder.set_supply_voltage(supply_voltage)

                sim_module.drive_motor.setInputVoltage(
                    self.add_friction(drive_motor.motor_voltage, sim_module.drive_friction_voltage)
                )
                sim_module.steer_motor.setInputVoltage(
                    self.add_friction(steer_motor.motor_voltage, sim_module.steer_friction_voltage)
                )

                sim_module.drive_motor.update(dt)
                sim_module.steer_motor.update(dt)

                drive_motor.set_raw_rotor_position(sim_module.drive_motor.getAngularPosition() / (2 * math.pi) * sim_module.drive_gearing)
                drive_motor.set_rotor_velocity(sim_module.drive_motor.getAngularVelocity() / (2 * math.pi) * sim_module.drive_gearing)

                steer_motor.set_raw_rotor_position(sim_module.steer_motor.getAngularPosition() / (2 * math.pi) * sim_module.steer_gearing)
                steer_motor.set_rotor_velocity(sim_module.steer_motor.getAngularVelocity() / (2 * math.pi) * sim_module.steer_gearing)

                # CANcoders see the mechanism, so don't account for the steer gearing
                cancoder.set_raw_position(sim_module.steer_motor.getAngularPosition() / (2 * math.pi))
                cancoder.set_velocity(sim_module.steer_motor.getAngularVelocity() / (2 * math.pi))

            states = [module.get_current_state() for module in modules_to_apply]

            angular_vel: radians_per_second = self._kinem.toChassisSpeeds(tuple(states)).omega
            self._last_angle = self._last_angle + Rotation2d(angular_vel * dt)
            self._pigeon_sim.set_raw_yaw(self._last_angle.degrees())
            self._pigeon_sim.set_angular_velocity_z(math.degrees(angular_vel))

        @staticmethod
        def add_friction(motor_voltage: volt, friction_voltage: volt) -> volt:
            """
            Applies the effects of friction to dampen the motor voltage.

            :param motor_voltage: Voltage output by the motor
            :type motor_voltage: volt
            :param friction_voltage: Voltage required to overcome friction
            :type friction_voltage: volt
            :returns: Friction-dampened motor voltage
            :rtype: volt
            """
            if math.fabs(motor_voltage) < friction_voltage:
                motor_voltage = 0.0
            elif motor_voltage > 0.0:
                motor_voltage -= friction_voltage
            else:
                motor_voltage += friction_voltage
            return motor_voltage

except ImportError:
    pass
